import grievous
import argparse
import logging


def main():
    logger = logging.getLogger()
    handler = logging.StreamHandler()

    parser = argparse.ArgumentParser()

    description = """
      _____ _____  _____ ________      ______  _    _  _____ 
     / ____|  __ \|_   _|  ____\ \    / / __ \| |  | |/ ____|
    | |  __| |__) | | | | |__   \ \  / / |  | | |  | | (___  
    | | |_ |  _  /  | | |  __|   \ \/ /| |  | | |  | |\___ \ 
    | |__| | | \ \ _| |_| |____   \  / | |__| | |__| |____) |
     \_____|_|  \_\_____|______|   \/   \____/ \____/|_____/

    Generalized Realignment of Innocuous and Essential Variants Otherwise Utilized as Skewed

    GRIEVOUS is a command-line tool for ensuring cross-cohort variant consistency. Specifically, 
    for any given cohort, GRIEVOUS will identify, orient, and realign all valid biallelic variants 
    from a chromosomal-level plink2 summary file (i.e., .pvar) or summary statistic file to 
    a user-specific variant database. Upon database alignment completion for all datasets of interest, 
    GRIEVOUS returns a comprehensive, consistently oriented and indexed feature set (across all 
    datasets), which can be used for the user's task of choice (e.g., ML training, validation, and
    testing; PRS construction and validation, etc.). 
    
    Succinctly, GRIEVOUS organizes the variant galaxy to the standards of the trade federation.
    """

    parser = argparse.ArgumentParser(description = description, formatter_class = argparse.RawDescriptionHelpFormatter)

    subparsers = parser.add_subparsers(help = "Must be either: 'realign', 'merge', 'intersect', 'list_dbs', 'records', 'delete_database' (or of course 'general')", required = True, dest = "operation")

    realignParser = subparsers.add_parser("realign", help = "Generate realignment files for genomic (i.e., .pvar) or summary statistic (.ssf) files.", formatter_class = argparse.ArgumentDefaultsHelpFormatter)
    realignParser.add_argument("--file", "-f", type = str, required = True, help = "Path to file to be grievous aligned.")
    realignParser.add_argument("--database", "-d", type = str, default = None, help = """User defined alias for a grievous database/dictionary. If None (default) grievous
                            will check internal settings to see if a single registered database is established, and if so will employ that for realignment. If more than 
                            one database is registered, the user must specify a database alias. Database aliases can be accessed with 'grievous list_dbs'.""")
    realignParser.add_argument("--file_type", "-t", type = str, help = "File type in {'pvar', 'ssf'} when automatic detection of file extension fails (i.e., when file does not end with .pvar or .ssf).")
    realignParser.add_argument("--write_path", "-w", type = str, default = None, help = """Directory to which grievous realigned files will be written. 
                            If none provided, grievous will write aligned files to a new directory GrievousAlignedFiles in the directory where --file (-f) was provided.""")
    realignParser.add_argument("--mapping", "-m", type = str, default = None, help = "Path to a mapping file which registers columns of file(s) to be aligned to the grievous standard.")
    realignParser.add_argument("--comment_characters", "-c", nargs = '+', type = str, default = [], help = """All comment characters at the header of a file to be skipped. Accepts multiple
                            arguments in a space separated manner. Comment characters including a # should be wrapped in quotes (e.g., '##').""")
    realignParser.add_argument("--verbose", "-v", action = "store_true", default = False, help = "Flag that increases the output from the realignment process.")
    realignParser.add_argument("--return_all", "-r", action = "store_true", default = False, 
                            help = "Whether to return all columns of the original pvar/ssf file. Default: False.")
    realignParser.add_argument("--shutoff_db_update", "-s", action = "store_true", default = False, 
                            help = """Flag for whether database/dictionary updates should not be performed. To ensure grievous correctness this flag should only (if ever) be used on the final cohort/ssf 
                            for all datasets attempting to be realigned.""")


    mergeParser = subparsers.add_parser("merge", help = "Merge CHR level reports to composite reports. For ssf files, will also merge all CHR level grievous formatted ssfs to a composite ssf. Called after realign.", formatter_class = argparse.ArgumentDefaultsHelpFormatter)
    mergeParser.add_argument("--aligned_files", "-a", type = str, required = True, help = "Path to a completed grievous realign write path/out directory.")
    mergeParser.add_argument("--update_id", "-u", action = "store_true", default = False, help = """Flag for whether to update the IDs of a merged SSF with the grievous realigned IDs. Without this flag, merged SSF will
                            retain the old SSF IDs under the ID column, and the file will be indexed with the grievous realigned IDs. Flag has no effect on .pvar realigned directories.""")


    intersectParser = subparsers.add_parser("intersect", help = "Given n grievous realigned cohorts, generates an intersected feature set (i.e., all biallelic SNPs that are common to all n cohorts). Called after merge for each cohort.", formatter_class = argparse.ArgumentDefaultsHelpFormatter)
    intersectParser.add_argument("--write_path", "-w", type = str, required = True, help = "Directory to which grievous intersect output file will be written.")
    intersectParser.add_argument("--cohorts", "-c", nargs = '+', type = str, required = True, help = "All grievous realigned and merged directories desired for intersection. Accepts multiple arguments in a space separated manner.")
    intersectParser.add_argument("--exclude_biallelic_duplicates", "-xd", action = "store_true", default = False, help = """Flag as to whether to omit grievous identified biallelic SNPs that were identified as duplicated in any given cohort. 
                                By default these SNPs are included/returned by grievous intersect.""")
    intersectParser.add_argument("--exclude_palindromic", "-xp", action = "store_true", default = False, help = """Flag as to whether to omit grievous identified biallelic SNPs that are also palindromic SNPs.
                                By default these SNPs are included/returned by grievous intersect.""")
    intersectParser.add_argument("--out_name", "-n", type = str, default = "Grievous_Intersecting_SNPs", help = "Name of file returned by grievous intersect. Default: Grievous_Intersecting_SNPs")


    recordParser = subparsers.add_parser("records", help = """Identify all cohorts aligned and updated to a given grievous database/dictionary. Returns whole database records or a chromosomal subset
                                        of the database records when --chr parameter is employed.""", formatter_class = argparse.ArgumentDefaultsHelpFormatter)
    recordParser.add_argument("--database", "-d", type = str, default = None, help = """Alias for an already existing grievous database/dictionary. If None (default) grievous
                            will check to see if a single registered database is established, and if so it will report records for this database.""") 
    recordParser.add_argument("--chr", "-c", nargs = '+', type = str, default = None, help = "Report database records for specific chromosomes.") 
    recordParser.add_argument("--out", "-o", type = str, default = None, help = "Output file path to write database records. Default: None (i.e., write to current stream such as console).")


    listDictParser = subparsers.add_parser("list_dbs", help = "List all user-specified databases/dictionaries in your grievous library.", formatter_class = argparse.ArgumentDefaultsHelpFormatter)


    deleteDatabaseParser = subparsers.add_parser("delete_database", help = "Permanently delete an existing grievous database, including all variant orientations and records.", formatter_class = argparse.ArgumentDefaultsHelpFormatter)
    deleteDatabaseParser.add_argument("--database", "-d", type = str, required = True, help = "Alias for an already existing grievous database/dictionary.")


    generalParser = subparsers.add_parser("general", help = "Summon the general... if you dare.", formatter_class = argparse.ArgumentDefaultsHelpFormatter)


    args = parser.parse_args()

    if args.operation == "realign":
        if args.verbose:
            formatter = logging.Formatter("%(asctime)s - %(message)s")
            handler.setFormatter(formatter)
            logger.setLevel(logging.INFO)
            
        else:
            logger.setLevel(logging.WARNING)

        logger.addHandler(handler) 

        grievous.processes.Realign(file = args.file, dictionary = args.database, write_path = args.write_path, mapping = args.mapping, 
                                file_type = args.file_type, return_all = args.return_all, comment_characters = args.comment_characters, 
                                shutoff_dict_update = args.shutoff_db_update)
        
    elif args.operation == "merge":
        logger.setLevel(logging.INFO)
        logger.addHandler(handler)

        grievous.processes.Merge(realignment_directory = args.aligned_files, updateIDs = args.update_id)

    elif args.operation == "intersect":
        logger.setLevel(logging.INFO)
        logger.addHandler(handler)

        grievous.processes.Intersect(write_path = args.write_path, cohorts = args.cohorts, exclude_biallelic_duplicates = args.exclude_biallelic_duplicates,
                                    omit_palindromic = args.exclude_palindromic, file_output_name = args.out_name)

    elif args.operation == "records":
        logger.setLevel(logging.INFO)
        logger.addHandler(handler)
        grievous.processes.utils.Records(dbAlias = args.database, chromosomes = args.chr, out = args.out)

    elif args.operation == "list_dbs":
        logger.setLevel(logging.INFO)
        logger.addHandler(handler)
        grievous.processes.utils.ListDicts()

    elif args.operation == "delete_database":
        logger.setLevel(logging.INFO)
        logger.addHandler(handler)
        grievous.processes.utils.DeleteDatabase(dbAlias = args.database)

    else:
        logger.setLevel(logging.INFO)
        logger.addHandler(handler) 
        grievous.processes.utils.GeneralGrievous()    

if __name__=="__main__":
	main()





import os 
import logging
import random

logger = logging.getLogger(__name__)

'''
About: Method to check if .grievous directory has been created, which serves as the grievous 
dictionary/database library. If it does not exist this method creates it. Returns the path
to the .grievous directory.

Input(s): None
Output(s): grievousDir: String corresponding to the .grievous directory which contains all databases.
'''
def _QueryGrievousInstall():
    homeDir = os.path.expanduser('~')
    grievousDir = os.path.join(homeDir, ".grievous")

    if not os.path.isdir(grievousDir):
        os.makedirs(grievousDir)
        assert os.path.isdir(grievousDir), f"Unable to create .grievous folder in {homeDir}. Exiting..."

    return grievousDir

'''
About: Method to display all user-created grievous databases.

Input(s): None
Output(s): None
'''
def ListDicts():
    grievousDir = _QueryGrievousInstall()

    #get database/dictionary aliases
    grievousLibrary = [el for el in os.listdir(grievousDir) if os.path.isdir(os.path.join(grievousDir, el))]

    logger.info("\nCurrent grievous database aliases:\n" + "-"*37)
    for dictionary in grievousLibrary:
        logger.info(dictionary)

    logger.info("")

    return None 

'''
About: Method to identify all palindromic SNPs.

Input(s): candidateSNPs: Iterable (i.e., set, list, tuple) of SNPs
Output(s): palindromicSNPs: set of all palindromic SNPs in candidate SNPs
'''
def _IdentifyPalindromic(candidateSNPs):
    palindromicPairs = {"A:T", "C:G"}
    palindromicSNPs = set()
    for snp in candidateSNPs:
        snpPair = ":".join(sorted(snp.split(":")[-2:]))
        if snpPair in palindromicPairs:
            palindromicSNPs.add(snp)

    return palindromicSNPs

'''
About: This method generates a record report for a given database. Specifically
it returns all cohorts/files that have been grievous realigned to a given database/dictionary
and employed for dictionary/database updates. This method returns records to logger 
(i.e., user stream) in the absence of out (i.e., out is None).

Input(s): 1) dbAlias: String corresponding to the grievous dictionary/database for which the user is attempting to access records.
          2) chromosomes: List of strings corresponding to a subset of particular chromosomes for which to access records (as opposed to all).
          3) out: String corresponding to a path (including file name and extension) to which to write files.

Output(s): None
Write(s):  recordReport - A report of all aligned and updated files aligned to a given database. If chromosomes is provided 
           then report is limited to those chromosomes of interest. Otherwise returns records for all chromosomes.
'''
def Records(dbAlias, chromosomes, out):
    if out: #Validate out path if user provided one:
        if os.path.dirname(out) != '': 
            assert os.path.isdir(os.path.dirname(out)), f"--out parent directory {os.path.abspath(os.path.dirname(out))} does not exist. Provide valid path to --out."

        if os.path.isdir(out): #user provided a directory path with no file name
            out = os.path.join(out, "GRIEVOUS_RECORDS.txt")

    grievousDir = _QueryGrievousInstall()

    #get database/dictionary aliases
    grievousLibrary = [el for el in os.listdir(grievousDir) if os.path.isdir(os.path.join(grievousDir, el))]

    #check database alias or if none provided then ensure only single db exists
    grievousRecordsDir = ""

    if dbAlias not in grievousLibrary:
        if dbAlias is None:
            assert len(grievousLibrary) == 1, """--database not provided and either more than one database exists or none have been created. 
            Provide a current database/dictionary alias to access corresponding records. Database aliases can be found with 'grievous list_dbs'."""

            grievousRecordsDir = os.path.join(grievousDir, grievousLibrary[0], "Records")
            dbAlias = grievousLibrary[0]
        
        else: #database alias provided by user does not exist
            raise Exception(f"--database {dbAlias} not found among grievous databases. Current grievous databases can be found with 'grievous list_dbs'.")

    else:
        grievousRecordsDir = os.path.join(grievousDir, dbAlias, "Records")

    chromosomalIteration = [str(i) for i in range(1,23)] + ["X", "Y", "MT"]
    
    if chromosomes: # user provided chromosomal subset
        #validate to ensure is a subset of chromosomalIteration
        for chromosome in chromosomes:
            assert chromosome in chromosomalIteration, f"--chr {chromosome} unknown. Valid --chr inputs are 1-22, X, Y, MT."

        chromosomalIteration = chromosomes

    #Add line at top of report indicating the database records reported
    if out:
        with open(out, "a") as recordReport:
            recordReport.write(f"Reporting Records For Database:  {dbAlias} \n\n")

    else:
        logging.info(f"Reporting Records For Database:  {dbAlias} \n")

    for chromosome in chromosomalIteration:
        with open(os.path.join(grievousRecordsDir, f"CHR_{chromosome}.aligned.record.txt"), "r") as alignedRecord:
            alignedContent = alignedRecord.readlines()

        with open(os.path.join(grievousRecordsDir, f"CHR_{chromosome}.updated.record.txt"), "r") as updatedRecord:
            updatedContent = updatedRecord.readlines()
        
        if out: #write to file if user provided one
            with open(out, "a") as recordReport:
                recordReport.write(f"CHR {chromosome} Records:\n\nALIGNED\n\n")
                for line in alignedContent:
                    recordReport.write(line)

                recordReport.write("\nUPDATED DATABASE:\n\n")

                for line in updatedContent:
                    recordReport.write(line)

                recordReport.write("\n")

        else:
            logger.info(f"CHR {chromosome} Records:\n\nALIGNED\n")
            for line in alignedContent:
                logger.info(line)

            logger.info("UPDATED DATABASE:\n") 

            for line in updatedContent:
                logger.info(line)
            

    return None

'''
About: This method takes in a database alias and if it exists deletes the database from the 
grievous library.

Input(s): dbAlias: String corresponding to the grievous dictionary/database the user is attempting to delete.
Output(s): None
'''
def DeleteDatabase(dbAlias):
    grievousDir = _QueryGrievousInstall()

    #get database/dictionary aliases
    grievousLibrary = [el for el in os.listdir(grievousDir) if os.path.isdir(os.path.join(grievousDir, el))]

    if dbAlias not in grievousLibrary:
        raise Exception(f"--database {dbAlias} not found among grievous databases. Current grievous databases can be found with 'grievous list_dbs'.")
    
    databaseToBeDeleted = os.path.join(grievousDir, dbAlias)
    recordsToRemove = os.path.join(databaseToBeDeleted, "Records")

    #delete records
    if os.path.isdir(recordsToRemove): #handle case of user interruption and entire Records is deleted, but not the entire database
        for record in os.listdir(recordsToRemove):
            toRemove = os.path.join(recordsToRemove, record)
            os.remove(toRemove)

        os.rmdir(recordsToRemove)

    #delete all chromosomal level orientation databases
    for chromDatabase in os.listdir(databaseToBeDeleted):
        toRemove = os.path.join(databaseToBeDeleted, chromDatabase)
        os.remove(toRemove)
    
    os.rmdir(databaseToBeDeleted)

    return None

'''
About: Every story needs a good old-fashioned villain... 
(Translation: What good is naming a tool grievous if you can't
have fun with it?). Queue Revenge of the SNP!

Input(s): None
Output(s): None
'''
def GeneralGrievous():
    quotes = ["If a variant does not appear in our records, it does not exist... yet.",
              "Your variants will make a fine addition to my collection!",
              "I will deal with this variant slime myself...",
              "I don't like SNPs. They're coarse and rough and irritating and they get everywhere.",
              "Wait a minute. How did this happen? We're smarter than this.\nApparently not.",
              "General Grievous, you're shorter than I expected.",
              "Hello there!",
              "Army or not, you must realize you are doomed!",
              "Only a SNP deals in absolutes. I will do what I must.",
              "SNP lords are our speciality.",
              "No giving up general jar jar. Weesa think of something.\nMy give up. My give up.",
              "Activate ray shields!",
              "Don't bother with them. Keep the SNPs in orbit!"]
    
    chosenOne = random.choice(quotes)

    grievousBanner = ["@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@","@@@@@@@@@@@@@@@@@@::..::@@@@@@@@@@@@@@@@@@",
                      "@@@@@@@@@@@@@@@@@:......::@@@@@@@@@@@@@@@@","@@@@@@@@@@@- @@@:.:......::@@ @@@@@@@@@@@@",
                      "@@@@@@@@@@@@ @@@:.=:....-::@@:@@@@@@@@@@@@","@@@@@@@@@@@@.@@-::=.....--:@@.@@@@@@@@@@@@",
                      "@@@@@@@@@@@. @@-.:-.....:=:=@  @@@@@@@@@@@","@@@@@@@@@@ ..@+-..=.....:::-@.  @@@@@@@@@@",
                      "@@@@@@@@@ ..:-#-..-.....:=--@..  @@@@@@@@@","@@@@@@@@ ..::--:.::.....:#::-::.  @@@@@@@@",
                      "@@@@@@@....::-=:.-:.:.:::#::+....  @@@@@@@","@@@@@@@...:*:-::.*..:...:#::--%:.. @@@@@@@",
                      "@@@@@@@:...@*-:::#::.::::#::::*#.  @@@@@@@","@@@@@@@@..+##-:::*:::::::#::--@@.. @@@@@@@",
                      "@@@@@@@@.:@#*:%@%@%:.::-*@%@%=@@:. @@@@@@@","@@@@@@@@:.-**:%@*+#*:::+@#*@%:#@..:@@@@@@@",
                      "@@@@@@@@:.:*#-%@@@#:::::*@@@%-*@..@@@@@@@@","@@@@@@@@..:=#==%@*::::::-=@%==*.. @@@@@@@@",
                      "@@@@@@@@=.::%#-----::::---=-:%+:. @@@@@@@@","@@@@@@@@@-:.-%=-==-:::----==-+:. .@@@@@@@@",
                      "@@@@@@@@@@-::=*++=-::::-===++-:::@@@@@@@@@","@@@@@@@@@@@::-+@%=-::::--=##=-::@@@@@@@@@@",
                      "@@@@@@@@@@@+::=%@#-::::--*@%-::@@@@@@@@@@@","@@@@@@@@@@@@-:-%@#-::::--*@#:::@@@@@@@@@@@",
                      "@@@@@@@@@@@@@::#@@-::::--%@+::@@@@@@@@@@@@","@@@@@@@@@@@@++::%@-:::--=%%::-@@@@@@@@@@@@",
                      "@@@@@@@@@@@+**:.%@-:----=@%.:%@@@@@@@@@@@@","@@@@@@@@@@*%*@=.@%-=*+===#@.-%%@@@@@@@@@@@",
                      "@@@@@@@@@++%%@@.#@-=*+==-@@.%@#+@@@@@@@@@@","@@@@@@@@-=**@@@-#@:+*++=-@%.@@##-@@@@@@@@@",
                      "@@@@@@@@==#%@@@%*%++***==##-@@@%=-@@@@@@@@","@@@@@@@==+%#@@@@:*:***++-*#@@@%%+=-@@@@@@@",
                      "@@@@@@@==#@%@@@@#=:++#+*-*@@@@#@%+-@@@@@@@","@@@@@@-=*%@@*@@@%=:=****-*@@@@%@@*-:@@@@@@",
                      "@@@@+*-=*@@@#@@%@+:*####-#@@@%@@@%-:*@@@@@"]

    logger.info("\n")
    for line in grievousBanner:
        logger.info(line)

    logger.info(f"\n\n{chosenOne}\n")

    return None


# @author: James V. Talwar

import pandas as pd
import os
import logging
from .pvar import Pvar
from ._utils import CleanReorientIndex

logger = logging.getLogger(__name__)

class SSF(Pvar):
    '''
    Parameters:
       filePath: String corresponding to where the file can be found (e.g., ssf)
       mappingPath: String corresponding to where column conversion standard can be found (default = None)
       commentCharsToSkip: List of strings corresponding to contiguous leading comments as are found across genomic files (default = ["##"]). 
                           
    '''
    def __init__(self, filePath, mappingPath = None, commentCharsToSkip = ["##"], *args, **kwargs):
        super().__init__(filePath, mappingPath, commentCharsToSkip, *args, **kwargs) 
        self.fileType = ".ssf"

        assert "BETA" in self.file.columns, "SSF Beta value missing (expecting BETA)"

    '''
    About: Writes all reports as well as generates a CHR-level formatted, oriented, and aligned SSF file needed for tasks such as SNP-set identification 
           and extraction by p-value in genotype cohorts or PRS construction. Note this CHR-level SSF employs the aligned GRIEVOUS standard IDs as the
           index and returns the original IDs as ID.
    
    Input(s):  1) writePath: String corresponding to the directory where want to write file(s).
               2) returnFullFile: Boolean corresponding to whether to return the full aligned ssf file (including
                  any uncorrected information such as INFO columns) or the grievous SSF column subset: [CHR, POS, ID, REF, ALT, BETA].
    
    Output(s): None 
    
    Write(s): 1) Chromosome-level grievous formatted and aligned SSF (as a .tsv) 
              2) A chromosome-level original ID SNP duplication report (if any duplicate IDs exist in the original file)
              3) The chromosome-level identified biallelic SNPs
              4) The chromosome-level subset of identified biallelic SNPs that were reoriented from cohort orientation to grievous dictionary standard (if these exist).
    ''' 
    def Write(self, writePath, returnFullFile):
        getChromosomeID = list(self.file.CHR)[0]
        
        if returnFullFile:
            self.file.to_csv(os.path.join(writePath, "GRIEVOUS_Formatted/FormattedAndAlignedSSF_CHR{}.tsv".format(getChromosomeID)), sep = "\t")
        
        else:
            self.file[["CHR", "POS", "ID", "REF", "ALT", "BETA"]].to_csv(os.path.join(writePath, "GRIEVOUS_Formatted/FormattedAndAlignedSSF_CHR{}.tsv".format(getChromosomeID)), sep = "\t")
    
        logger.info("Finished writing aligned and formatted chromosome {} ssf\n".format(getChromosomeID))

        #Identify and report ID duplicates in the SSF if they exist
        if sum(self.file.duplicated(subset = ["ID"], keep = False)) > 0:
            logger.info(f"Original CHR {self.Chrom} SSF file has duplicate IDs - generating duplication report.")

            oldIDToFormattedID = self.file[["ID"]].copy()
            oldIDToFormattedID["ProperlyFormattedName"] = oldIDToFormattedID.index
            oldIDToFormattedID.index = [i for i in range(len(oldIDToFormattedID.index))]

            _, duplicatedIDReport, duplicatedIDCounts = CleanReorientIndex(oldIDToFormattedID, self.biallelicSNPs)
            
            with open(os.path.join(writePath, "Reports/CHR{}_Duplication_Report.txt".format(getChromosomeID)), "a") as duplicationReport:
                duplicationReport.write("Original file duplication ID incidences:\n\n")
                
                for k,v in duplicatedIDCounts.items():
                    duplicationReport.write("\t".join([k,str(v)]) + "\n")
                
                duplicationReport.write("\nOriginal ID to grievous index unique incidence mappings:\n")

                for k,v in duplicatedIDReport.items():
                    duplicationReport.write("\n" + k + "\n\n")
                    for originalID, grievousIndexes in v.items():
                        duplicationReport.write("\t".join([originalID , str(grievousIndexes)]) + "\n")

        logger.info("Writing biallelic SNPs and subset of biallelic SNPs that were reoriented from cohort specific orientation to grievous database standard\n")
        
        pd.DataFrame(self.biallelicSNPs, columns = ["BiallelicSNPs"]).to_csv(os.path.join(writePath, "Reports/CHR{}_BiallelicSNPs.tsv".format(getChromosomeID)), sep = "\t", index = False)
        if len(self.flippedSNPs) > 0:
            pd.DataFrame(self.flippedSNPs, columns = ["FlippedSNPs"]).to_csv(os.path.join(writePath, "Reports/CHR{}_FlippedSNPs.tsv".format(getChromosomeID)), sep = "\t", index = False)

        logger.info("GRIEVOUS Alignment for CHR {} complete\n".format(getChromosomeID))
        
        return None



#@author: James V. Talwar

import os
import pandas as pd
import logging
import re

logger = logging.getLogger(__name__)


'''
About: This method takes in a path to a grievous realigned write directory and generates composite reports across
all chromosomes that were completed at the time merge was called. If the write directory passed in was for an ssf
file, merge will also generate a composite ssf in the GRIEVOUS_Formatted subfolder.

Input(s):  1) realignment_directory: String corresponding to the --write_path (-w) of a completed grievous realign.
           2) updatedIDs: Boolean corresponding to whether to update/replace the original SSF IDs with the greivous realigned indexes.
Output(s): None
Write(s):  Composite Biallelic and Flipped SNP reports for a given cohort and a composite ssf if the realignmentDirectory was
           for an ssf file.
'''
def Merge(realignment_directory, updateIDs):
    #check if is a valid realignment directory
    assert os.path.isdir(realignment_directory), f"""Given path {realignment_directory} for --aligned_files does not exist/is not a directory. Please retry with a valid directory
    that is the output of a 'grievous realign' run."""

    subdirectories = os.listdir(realignment_directory)
    assert ("Reports" in subdirectories) and ('GRIEVOUS_Formatted' in subdirectories), f"Unable to locate both Reports and GRIEVOUS_Formatted folders in {realignment_directory}. Confirm directory - recall 'grievous realign' must be called before 'grievous merge'."

    #Check the file extension for ssfs - ensure unique (i.e., user did not reuse same directory for different realigns)
    grievousFormattedDir = os.path.join(realignment_directory, "GRIEVOUS_Formatted")
    fileType = set([el.split("_")[0].split("FormattedAndAligned")[1] for el in os.listdir(grievousFormattedDir) if el.endswith(".tsv")])

    assert (len(fileType) == 1) and ({"Pvar", "SSF"}.intersection(fileType) == fileType), f"{len(fileType)} file types found in --aligned_files directory {realignment_directory}. Given directory has realignment extensions: {fileType}. Ensure that every dataset-level grievous realign has a unique write_path."
    
    extension = next(iter(fileType))

    #grab all the biallelic SNPs and flipped SNPs and merge 
    reportDir = os.path.join(realignment_directory, "Reports")
    reportDirFiles = os.listdir(reportDir)

    if ("ALL_BiallelicSNPs.tsv" in reportDirFiles) or ("ALL_FlippedSNPs.tsv" in reportDirFiles):
        logger.warning(f"WARNING: Detected previous merge files ALL_BiallelicSNPs/ALL_FlippedSNPs.tsv in --aligned_files {os.path.join(realignment_directory, 'Reports')}. Overwriting...\n")
    
    mergeReportFiles = [os.path.join(reportDir, file) for file in reportDirFiles if (file.endswith(".tsv")) and (not re.match("WARNING_CHR.*_BiallelicDuplicates.tsv", file)) and (file not in {"ALL_FlippedSNPs.tsv", "ALL_BiallelicSNPs.tsv"})]

    biallelicSNPs = list()
    flippedSNPs = list()

    for report in mergeReportFiles:
        prefix = report.split("/")[-1].split("_")[-1].split(".")[0]
        assert prefix in {"BiallelicSNPs", "FlippedSNPs"}, f"Unknown report {report} with prefix {prefix} found. Valid prefixes for reports are 'BiallelicSNPs' and 'FlippedSNPs'. Exiting..."

        toAppend = pd.read_csv(report, sep = "\t")[prefix].tolist()

        if prefix == "BiallelicSNPs":
            biallelicSNPs = biallelicSNPs + toAppend

        else:
            flippedSNPs = flippedSNPs + toAppend


    #Write AllBiallelic and AllFlipped files
    if len(biallelicSNPs) > 0:
        pd.DataFrame(biallelicSNPs, columns = ["BiallelicSNPs"]).to_csv(os.path.join(reportDir, "ALL_BiallelicSNPs.tsv"), sep = "\t", index = False)

    if len(flippedSNPs) > 0:
        pd.DataFrame(flippedSNPs, columns = ["FlippedSNPs"]).to_csv(os.path.join(reportDir, "ALL_FlippedSNPs.tsv"), sep = "\t", index = False)

    logger.info("Biallelic and flipped SNP reports merged.")

    #if SSF merge chromosomal-level SSFs into one cohesive SSF
    if extension == "SSF":
        logger.info("\nMerging FormattedAndAlignedSSF files")
        mergeSSFFiles = [os.path.join(grievousFormattedDir, file) for file in os.listdir(grievousFormattedDir) if file.endswith(".tsv")]

        mergedSSF = pd.DataFrame()
        for ssf in mergeSSFFiles:
            chrSSF = pd.read_csv(ssf, sep = "\t", index_col = 0) 
            mergedSSF = pd.concat([mergedSSF, chrSSF], axis = 0)

        if updateIDs:
            mergedSSF.ID = mergedSSF.index
            mergedSSF.to_csv(os.path.join(grievousFormattedDir, "MergedSSF.ssf"), sep = "\t", index = False)
        
        else:
            mergedSSF.to_csv(os.path.join(grievousFormattedDir, "MergedSSF.ssf"), sep = "\t")

        
    logger.info("\ngrievous merge complete!")


    return None





#@author: James V. Talwar

import os
import pandas as pd
import logging
import re
from .utils import _IdentifyPalindromic

logger = logging.getLogger(__name__)

'''
About: This method takes in n number of grievous realigned and merged cohorts and generates the feature/SNP set 
common to all cohorts of interest.

Input(s):  1) write_path: String corresponding to the directory to which write the grievous intersect feature set.
           2) cohorts: List of strings corresponding to the directories of grievous realigned and merged feature sets. 
           For each cohort this corresponds to the --write_path provided to grievous realign and to the --aligned_files 
           provided to grievous merge.
           3) exclude_biallelic_duplicates: Boolean corresponding to whether to exclude or include biallelic SNPs that were 
           originally duplicated in a cohort.
           4) omit_palindromic: Boolean corresponding to whether to omit palindromic SNPs from the intersected feature set.
           5) file_output_name: String corresponding to the output file name written by grievous intersect.    
Output(s): None
Write(s):  An intersected SNP/feature set common to all cohorts provided in directory write_path with file name
           file_output_name.
'''

def Intersect(write_path, cohorts, exclude_biallelic_duplicates, omit_palindromic, file_output_name):
    logger.info(f"Cohorts for intersection: {cohorts}\n")

    #Validate user inputs:
    assert os.path.isdir(write_path), f"--write_path {write_path} does not exist as a directory. Please retry with a valid write_path. Exiting..."
    
    for cohort in cohorts:
        assert os.path.isdir(cohort), f"Provided cohort {cohort} does not exist as a directory. Exiting..."

        subdirectories = os.listdir(cohort)
        assert ("Reports" in subdirectories) and ('GRIEVOUS_Formatted' in subdirectories), f"""Unable to locate both Reports and GRIEVOUS_Formatted folders in cohort {cohort}.
          Confirm directory and recall 'grievous intersect' is called after a given cohort has undergone both 'grievous realign' and 'grievous merge'."""
        
        assert "ALL_BiallelicSNPs.tsv" in os.listdir(os.path.join(cohort, "Reports")), f"""ALL_BiallelicSNPs.tsv file missing from cohort {cohort}. Validate 'grievous merge' was
        run. If so, then no biallelic SNPs exist for cohort {cohort} and the intersection across all cohorts of interest is an empty set."""

    intersectingSNPs = set()
    firstInstance = True

    for cohort in cohorts:
        reportDir = os.path.join(cohort, "Reports")
        cohortAlignedBiallelicSNPs = set(pd.read_csv(os.path.join(reportDir, "ALL_BiallelicSNPs.tsv"), sep = "\t")["BiallelicSNPs"])

        if exclude_biallelic_duplicates:
            duplicatesToOmit = set()
            cohortBiallelicDuplicateFiles = [os.path.join(reportDir, file) for file in os.listdir(reportDir) if re.match("WARNING_CHR.*_BiallelicDuplicates.tsv", file)]
            for dups in cohortBiallelicDuplicateFiles:
                 duplicatesToOmit = duplicatesToOmit.union(set(pd.read_csv(dups, sep = "\t", header = None)[0]))

            cohortAlignedBiallelicSNPs = cohortAlignedBiallelicSNPs.difference(duplicatesToOmit)

        if firstInstance: # base case (i.e., first cohort in cohorts)
            firstInstance = False
            
            #Handle corner case where all SNPs are duplicated and exclude_biallelic_duplicates called - write an empty intersection (no SNPs to intersect)
            if len(cohortAlignedBiallelicSNPs) == 0:
                break

            intersectingSNPs = cohortAlignedBiallelicSNPs

        else:
            intersectingSNPs = intersectingSNPs.intersection(cohortAlignedBiallelicSNPs)


    #Handle palindromic SNP exclusion:
    if omit_palindromic:
        logger.info("\nExcluding palindromic SNPs...\n")
        palindromicSNPs = _IdentifyPalindromic(intersectingSNPs)
        intersectingSNPs = intersectingSNPs.difference(palindromicSNPs)
        
        logger.info(f"\nIdentified and removed {len(palindromicSNPs)} palindromic SNPs.\n")

    logger.info(f"The number of intersecting SNPs across the given {len(cohorts)} cohorts is: {len(intersectingSNPs)}. Writing intersecting SNPs...")

    pd.DataFrame(intersectingSNPs).to_csv(os.path.join(write_path, file_output_name + ".tsv"), sep = "\t", index = False, header = None)

    logger.info(f"\nIntersecting SNPs/features identified and written to {os.path.join(write_path, file_output_name + '.tsv')}. My work here is done...")

    return None


from setuptools import setup, find_packages

setup(
    name = "grievous",
    version = "0.1.0",
    author = "James V. Talwar",
    author_email = "jtalwar@ucsd.edu",
    description = "Generalized Realignment of Innocuous and Essential Variants Otherwise Utilized as Skewed",
    license = "LICENSE",
    url = "https://github.com/jvtalwar/GRIEVOUS", 
    packages = find_packages(),
    python_requires = '>=3.7',
    install_requires = ["pandas >= 1.3.4"],

    scripts=["grievous/generalGrievous.py"],

    entry_points = {
        "console_scripts": [
            "grievous = grievous.generalGrievous:main"
        ]
    }

)